﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Data;
using System.Windows.Forms;

namespace RikaiChanClone
{
    static class JEDictionary
    {
        static IDbConnection connection = null;
        static bool ready = false;

        static void Init()
        {
            var sb = new System.Data.SQLite.SQLiteConnectionStringBuilder();
            var dbPath = FindDictionary();

            if (dbPath != null)
            {
                sb.DataSource = dbPath;
                connection = new System.Data.SQLite.SQLiteConnection(sb.ToString());
            }

            ready = true;
        }


        static string FindDictionary()
        {
            string fileName = "dict.sqlite";
            string testPath;
            testPath = Application.StartupPath + fileName;
            if (File.Exists(testPath)) return testPath;
            testPath = Environment.CurrentDirectory + fileName;
            if (File.Exists(testPath)) return testPath;

            testPath = ScanApplicationData();
            if (!String.IsNullOrEmpty(testPath)) return testPath;

            return null;
        }

        static private string ScanApplicationData()
        {
            var appData = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
            var profileBase = appData + "/mozilla/firefox/profiles/";
            var profiles = GetDirectories(profileBase);
            if (profiles != null)
            {
                foreach (var profile in profiles)
                {
                    var extensionbase = profile + "/extensions/";
                    var extensions = GetDirectories(extensionbase);
                    if (extensions == null) break;
                    foreach (var extension in extensions)
                    {
                        string testPath = extension + "/dict.sqlite";
                        if (File.Exists(testPath))
                        {
                            return testPath;
                        }
                    }
                }
            }
            return null;
        }

        private static string[] GetDirectories(string profileBase)
        {
            try
            {
                var profiles = Directory.GetDirectories(profileBase);
                return profiles;
            }
            catch (IOException)
            {
                return null;
            }
            catch (UnauthorizedAccessException)
            {
                return null;
            }
        }

        public static List<DictionaryEntry> LookupKanji(string text)
        {
            if (!ready)
            {
                Init();
            }
            if (connection == null) return null;


            var command = connection.CreateCommand();
            command.CommandText = "SELECT kanji,kana,entry FROM dict WHERE kanji = @kanji";
            command.CommandType = CommandType.Text;
            var param = command.CreateParameter();
            param.ParameterName = "@kanji";
            param.Value = text;
            command.Parameters.Add(param);
            if (connection.State != ConnectionState.Open)
            {
                connection.Open();
            }
            var reader = command.ExecuteReader();

            List<DictionaryEntry> result = null;

            while (reader.Read())
            {
                if (result == null) result = new List<DictionaryEntry>();
                DictionaryEntry entry = new DictionaryEntry();

                entry.TextLength = text.Length;
                if (!reader.IsDBNull(0))
                {
                    entry.Kanji = reader.GetString(0);
                }
                if (!reader.IsDBNull(1))
                {
                    entry.Kana = reader.GetString(1);
                }
                if (!reader.IsDBNull(2))
                {
                    entry.Entry = reader.GetString(2);
                }
                result.Add(entry);
            }
            return result;
        }

        public static List<DictionaryEntry> LookupKana(string text)
        {
            if (!ready)
            {
                Init();
            }
            if (connection == null) return null;

            
            var command = connection.CreateCommand();
            command.CommandText = "SELECT kanji,kana,entry FROM dict WHERE kana = @kana";
            command.CommandType = CommandType.Text;
            var param = command.CreateParameter();
            param.ParameterName = "@kana";
            param.Value = text;
            command.Parameters.Add(param);

            if (connection.State != ConnectionState.Open)
            {
                connection.Open();
            }

            var reader = command.ExecuteReader();

            List<DictionaryEntry> result = null;

            while (reader.Read())
            {
                if (result == null) result = new List<DictionaryEntry>();
                DictionaryEntry entry = new DictionaryEntry();

                entry.TextLength = text.Length;
                if (!reader.IsDBNull(0))
                {
                    entry.Kanji = reader.GetString(0);
                }
                if (!reader.IsDBNull(1))
                {
                    entry.Kana = reader.GetString(1);
                }
                if (!reader.IsDBNull(2))
                {
                    entry.Entry = reader.GetString(2);
                }
                result.Add(entry);
            }
            return result;
        }
        public static List<DictionaryEntry> Lookup(string word)
        {
            List<DictionaryEntry> list = new List<DictionaryEntry>();
            word = word.ToHiragana();
            if (word.Length == 0) return list;

            int MaxLength = 16;
            MaxLength = Math.Min(word.Length, MaxLength);

            char c = word[0];
            if (c >= 0x3040 && c <= 0x30FF)  //only kana
            {
                for (int l = MaxLength; l > 0; l--)
                {
                    var results = LookupKana(word.Substring(0, l));
                    if (results != null)
                    {
                        list.AddRange(results);
                    }
                }
            }
            else  //kanji
            {
                for (int l = MaxLength; l > 0; l--)
                {
                    var results = LookupKanji(word.Substring(0, l));
                    if (results != null)
                    {
                        list.AddRange(results);
                    }
                }
            }
            return list;

        }

        public static string LookupAll(string word)
        {
            var entries = Lookup(word);
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < entries.Count && i < 10; i++)
            {
                var entry = entries[i];
                sb.Append(entry.Kanji + "   " + entry.Kana + "   " + entry.Romaji);
                sb.AppendLine();
                sb.AppendLine(entry.Entry);
            }
            return sb.ToString();
        }
    }
}
