﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.IO;

namespace RikaiChanClone
{
    public partial class TextEditorForm : Form
    {
        const int SHIFTJIS = 932;
        const int WESTERN = 1252;
        
        string _loadedFileName;
        bool _modified;
        bool _saveAsUnix;
        bool _saveAsShiftJis;
        const string appTitle = "RikaiChan-clone Text Editor";

        string loadedFileName
        {
            get
            {
                return _loadedFileName;
            }
            set
            {
                _loadedFileName = value;
                if (value == null)
                {
                    this.Text = appTitle;
                }
                else
                {
                    this.Text = appTitle + " - " + value;
                }
            }
        }
        bool saveAsUnix
        {
            get
            {
                return _saveAsUnix;
            }
            set
            {
                _saveAsUnix = value;
            }
        }
        bool saveAsShiftJis
        {
            get
            {
                return _saveAsShiftJis;
            }
            set
            {
                _saveAsShiftJis = value;
            }
        }
        bool modified
        {
            get
            {
                return _modified;
            }
            set
            {
                _modified = value;
            }
        }


        public TextEditorForm()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            this.Text = appTitle;
        }

        private void textBox1_TextChanged(object sender, EventArgs e)
        {
            modified = true;
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            NewFile();
        }

        private void NewFile()
        {
            if (modified)
            {
                var dl = MessageBox.Show("File is changed, save first?", "Confirm New File", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (dl == DialogResult.Yes)
                {
                    if (!Save()) return;
                }
                if (dl == DialogResult.Cancel) return;
            }
            textBox1.Text = "";
            saveAsShiftJis = false;
            saveAsUnix = false;
            modified = false;
            loadedFileName = null;
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFile();
        }

        private bool OpenFile()
        {
            if (modified)
            {
                var dl = MessageBox.Show("File is changed, save first?", "Confirm Open File", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (dl == DialogResult.Yes)
                {
                    if (!Save()) return false;
                }
                if (dl == DialogResult.Cancel) return false;
            }

            var dialog = new OpenFileDialog();
            dialog.Filter = "Text Files (*.txt)|*.txt|All Files (*.*)|*.*";
            if (dialog.ShowDialog() == DialogResult.OK)
            {
                string fileName = dialog.FileName;
                return OpenFile(fileName);
            }
            return false;
        }

        private bool OpenFile(string fileName)
        {
            byte[] data;
            string text;
            try
            {
                data = File.ReadAllBytes(fileName);
                textBox1.Text = File.ReadAllText(fileName);
            }
            catch (IOException ex)
            {
                MessageBox.Show("IO Exception when opening file: " + ex.Message);
                return false;
            }

            try
            {
                var utf8WithExceptions = new UTF8Encoding(false, true);
                text = utf8WithExceptions.GetString(data);
                saveAsShiftJis = false;

            }
            catch (System.Text.DecoderFallbackException ex)
            {
                var ShiftJis = Encoding.GetEncoding(SHIFTJIS);
                text = ShiftJis.GetString(data);
                saveAsShiftJis = true;
            }

            if (text.Contains("\n") && !text.Contains("\r\n"))
            {
                text = text.Replace("\n", "\r\n");
                saveAsUnix = true;
            }
            else
            {
                saveAsUnix = false;
            }
            textBox1.Text = text;

            loadedFileName = fileName;
            modified = false;
            return true;
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Save();
        }

        private bool Save()
        {
            if (loadedFileName == null)
            {
                return SaveAs();
            }
            else
            {
                return SaveFile(loadedFileName);
            }
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveAs();
        }

        private bool SaveAs()
        {
            var dialog = new SaveFileDialog();
            dialog.Filter = "Text Files (*.txt)|*.txt|All Files (*.*)|*.*";
            if (loadedFileName != null)
            {
                dialog.FileName = loadedFileName;
            }

            if (dialog.ShowDialog() == DialogResult.OK)
            {
                string fileName = dialog.FileName;
                return SaveFile(fileName);
            }
            return false;
        }

        private bool SaveFile(string fileName)
        {
            var UTF8NoBom = new UTF8Encoding(false);
            var ShiftJis = Encoding.GetEncoding(SHIFTJIS);

            string text = textBox1.Text;
            if (saveAsUnix)
            {
                text = text.Replace("\r\n", "\n");
            }
            try
            {
                if (saveAsShiftJis == false)
                {
                    File.WriteAllText(fileName, text, UTF8NoBom);
                }
                else
                {
                    File.WriteAllText(fileName, text, ShiftJis);
                }
            }
            catch (IOException ex)
            {
                MessageBox.Show("IO Exception when saving file: " + ex.Message);
                return false;
            }

            loadedFileName = fileName;
            modified = false;

            return true;
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void useUNIXLineEndingsToolStripMenuItem_CheckStateChanged(object sender, EventArgs e)
        {
            var stripItem = (ToolStripMenuItem)sender;
            saveAsUnix = stripItem.Checked;
        }

        private void undoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.UndoRedo.Undo();
        }

        private void editToolStripMenuItem_DropDownOpening(object sender, EventArgs e)
        {
            undoToolStripMenuItem.Enabled = textBox1.UndoRedo.CanUndo;
            redoToolStripMenuItem.Enabled = textBox1.UndoRedo.CanRedo;
            cutToolStripMenuItem.Enabled = textBox1.Selection.Length > 0;
            copyToolStripMenuItem.Enabled = textBox1.Selection.Length > 0;
            pasteToolStripMenuItem.Enabled = Clipboard.ContainsText();
            selectAllToolStripMenuItem.Enabled = textBox1.TextLength > 0;
            useUNIXLineEndingsToolStripMenuItem.Checked = saveAsUnix;
            useShiftJISEncodingToolStripMenuItem.Checked = saveAsShiftJis;
        }

        private void cutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.Clipboard.Cut();
        }

        private void copyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.Clipboard.Copy();
        }

        private void pasteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.Clipboard.Paste();
        }

        private void selectAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.Selection.SelectAll();
        }

        private void zoomInToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.Zoom++;
        }

        private void zoomOutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.Zoom--;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (modified)
            {
                var dl = MessageBox.Show("File is changed, save first?", "Confirm Exit", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
                if (dl == DialogResult.Yes)
                {
                    if (!Save()) e.Cancel = true;
                }
                if (dl == DialogResult.Cancel)
                {
                    e.Cancel = true;
                }
            }

        }

        private void selectFontToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var fd = new FontDialog();
            fd.Font = textBox1.Font;
            if (fd.ShowDialog() == DialogResult.OK)
            {
                textBox1.Font = fd.Font;
                textBox1.ForEachStyle(s=>
                    {
                        s.Size = fd.Font.Size;
                        s.Font = fd.Font;
                    });
            }
        }

        private void wordWrapToolStripMenuItem_CheckStateChanged(object sender, EventArgs e)
        {
            if (wordWrapToolStripMenuItem.Checked)
            {
                textBox1.LineWrap.Mode = ScintillaNet.WrapMode.Word;
            }
            else
            {
                textBox1.LineWrap.Mode = ScintillaNet.WrapMode.None;
            }
        }

        private void Form1_DragDrop(object sender, DragEventArgs e)
        {
            string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);

            if (files.Length > 0)
            {
                OpenFile(files[0]);
            }
        }

        private void Form1_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop, false) == true)
            {
                e.Effect = DragDropEffects.All;
            }
        }

        private void findToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.FindReplace.ShowFind();
        }

        private void replaceToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.FindReplace.ShowReplace();
        }

        private void redoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.UndoRedo.Redo();
        }

        private void useShiftJISEncodingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var stripItem = (ToolStripMenuItem)sender;
            saveAsShiftJis = stripItem.Checked;
        }

        private void fixBrokenTextToolStripMenuItem_Click(object sender, EventArgs e)
        {
            textBox1.Text = FixBrokenText(textBox1.Text);
        }



        Dictionary<char, byte> _inverseWestern = null;
        Dictionary<char, byte> InverseWestern
        {
            get
            {
                if (_inverseWestern != null)
                {
                    return _inverseWestern;
                }
                _inverseWestern = new Dictionary<char, byte>();
                var Western = Encoding.GetEncoding(WESTERN);
                var text = Western.GetString(0.Until(256).Select(i => (byte)i).ToArray());
                text.ForEach((c, i) => _inverseWestern.Add(c, (byte)i));
                return _inverseWestern;
            }
        }

        int IndexOfNonWesternCharacter(string str, int start)
        {
            var inverseWestern = InverseWestern;
            int i;
            for (i = start; i < str.Length; i++)
            {
                if (!inverseWestern.ContainsKey(str[i]))
                {
                    return i;
                }
            }
            return i;
        }

        int IndexOfWesternCharacter(string str, int start)
        {
            var inverseWestern = InverseWestern;
            int i;
            for (i = start; i < str.Length; i++)
            {
                if (inverseWestern.ContainsKey(str[i]))
                {
                    return i;
                }
            }
            return i;
        }


        private string FixBrokenText(string text)
        {
            var Western = Encoding.GetEncoding(WESTERN);
            var ShiftJis = Encoding.GetEncoding(SHIFTJIS);
            var chars = text.ToCharArray();
            StringBuilder sb = new StringBuilder();
            int i=0;
            while (i < text.Length)
            {
                {
                    int next = IndexOfNonWesternCharacter(text, i);
                    int length = next - i;
                    var bytes = Western.GetBytes(chars, i, length);
                    var newText = ShiftJis.GetString(bytes);
                    sb.Append(newText);
                    i = next;
                }
                {
                    int next = IndexOfWesternCharacter(text, i);
                    int length = next - i;
                    var newText = text.Substring(i, length);
                    sb.Append(newText);
                    i = next;
                }
            }
            return sb.ToString();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            using (var form = new AboutBox1())
            {
                form.ShowDialog();
            }
        }
    }
}
